/*
   This file is part of the index package,
   Copyright (C) 1997 Harold Tay, haroldtay@pacific.net.sg

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA
 */
/*
   Compress the integer arrays.

   Each entry in the words section of the index contains the
   following fields, each of which is variable in length:

   word size off ...

   "word" is a string, so it's terminated with a \0.
   "size" is the size of the rest of the entry, in bytes.
   "off" are all the offsets, and there may be any number of them (but not 0).

   Each of the values after "word" is nominally an unsigned integer,
   but many of them will not need the full 32 bits.

   Each of these nominal integer values is represented by up to 4
   bytes, but the high bit of each byte is stolen to indicate the
   presence of -more- bytes.  So each nominal integer can contain a
   maximum of 29 bits, the last high bit is not stolen.  The byte
   order is lsb first.

   Each of the exported functions returns the length of the char
   buffer consumed or created; the first arg is the in param; if it
   is int *, then the next param is the length of it.  The next
   param is the out param; if it is int **, then the next param is
   an out param which is int *, the length of it.

   Observation: integers in the array increase monotonically, so
   just keep the difference.

   All return buffers should not be free()d, and will be overwritten
   on the next call.
 */

#include <stdio.h>
#include <stdlib.h>

/* Decode the buffer, write the decoded integer into
   the space provided and return the number of bytes consumed */

static inline int
readvint(char * buf, unsigned int * where)
{
  unsigned int t;
  unsigned char * bufp = (unsigned char *)buf;

  t = *bufp & 0x7f;
  if( *bufp & (1<<7) ){
    bufp++;
    t |= ((*bufp & 0x7f) << 7);
    if( *bufp & (1<<7) ){
      bufp++;
      t |= ((*bufp & 0x7f) << 14);
      if( *bufp & (1<<7) ){
        bufp++;
        t |= ((*bufp & 0xff) << 21);
      }
    }
  }
  if( where )*where = t;
  return(1 + bufp - (unsigned char *)buf);
}

/* Write the encoded integer into the given buffer space, return
   the number of bytes written */

static inline int
makevint(char * buf, register unsigned int t)
{
  register unsigned char * bufp = (unsigned char *)buf;

  *bufp = t & 0x7f;
  if( (t >>= 7) ){
    *bufp++ |= (1<<7); *bufp = t & 0x7f;
    if( (t >>= 7) ){
      *bufp++ |= (1<<7); *bufp = t & 0x7f;
      if( (t >>= 7) ){
        *bufp++ |= (1<<7); *bufp = t & 0xff;
      }
    }
  }
  return(1 + bufp - (unsigned char *)buf);
}

/* Create the compressed buffer.  alen is length of the array in ints. */

int
codec_co(int * a, int alen, char ** b)
{
  static char * buf = NULL;
  static int buf_size = 0;
  char s[4];
  char * bufp;
  int size;
  int i;

  if( buf_size < (alen + 1)*4 )
    buf = realloc(buf, buf_size = (alen + 1)*4);

  bufp = buf + 4;
  {
    unsigned int t = 0;
    for(i = 0; i < alen; t = a[i], i++)
      bufp += makevint(bufp, (unsigned int)a[i] - t);
  }

  size = makevint(s, (bufp - buf) - 4);
  for(i = 1; i <= size; i++)buf[4-i] = s[size-i];

  *b = buf + (4-size);
  return((bufp - buf) - (4-size));
}

/* Decompress the buffer into the int array. */

int
codec_dec(char * buf, int ** a, int * lenp)
{
  static unsigned int * ary = NULL;
  static int ary_size = 0;
  char * bufp = buf;
  char * end;
  unsigned int len;
  int i;

  bufp += readvint(bufp, &len);
  end = bufp + len;

  if( ary_size < len )
    ary = realloc(ary, (ary_size = len + 1) * sizeof(int));

  {
    unsigned int t = 0;
    for(i = 0; bufp < end; t = ary[i], i++){
      bufp += readvint(bufp, &ary[i]);
      ary[i] += t;
    }
  }

  *lenp = i;
  *a = ary;
  return(bufp - buf);
}

int
codec_arraylen(char * buf, int * arlen)
{
  char * bufp;
  unsigned int t;
  int skip;
  int len;

  skip = readvint(buf, &t);
  bufp = buf + skip;

  for(len = 0; bufp < buf + t; len++)
    bufp += readvint(bufp, NULL);
  *arlen = len;
  return(skip + t);
}

int
codec_skip(char * buf)
{
  unsigned int t;
  int skip;

  skip = readvint(buf, &t);
  return(skip + t);
}

